/*
 * MIT License
 * 
 * Copyright (c) 2019 Chirag Sakhuja
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
*/
#include "mypredictor.h"

/*
 * Much of this code is borrowed from the EVES CVP submission with the following modifications:
 *   - Code specific to VTAGE has been removed.
 *   - Formatting and organization has been minorly changed.
 * The code added for HCVP is clearly delimited with comments. Most of the logic to implement
 * HCVP is in mypredictor.h.
 */

int seq_commit;

#define NOTLLCMISS (actual_latency < 150)
#define NOTL2MISS (actual_latency < 60)
#define NOTL1MISS (actual_latency < 12)
#define FASTINST (actual_latency ==1)
#define MFASTINST (actual_latency <3)

/** BEGIN FOR HCVP **/
std::unordered_map<uint64_t, InstInfo> seq_no_to_info;
ValueCorrelationPredictor VP;
/** END FOR HCVP **/

void getPredStride(ForUpdate * U, uint64_t & predicted_value, uint64_t seq_no);
bool strideupdateconf (ForUpdate * U, uint64_t actual_value, int actual_latency, int stride);
bool StrideAllocateOrNot (ForUpdate * U, uint64_t actual_value, int actual_latency);
void UpdateStridePred (ForUpdate * U, uint64_t actual_value, int actual_latency);

bool getPrediction (uint64_t seq_no, uint64_t pc, uint8_t piece, uint64_t & predicted_value)
{
    /** BEGIN FOR HCVP **/
    InstInfo inst_info { seq_no, pc, piece };
    seq_no_to_info[seq_no] = inst_info;
    /** END FOR HCVP **/

    ForUpdate *U;
    U = &Update[seq_no & (MAXINFLIGHT - 1)];
    U->pc = pc + piece;
    U->predstride = false;
    U->predhcvp = false;

    getPredStride (U, predicted_value, seq_no);

    /** BEGIN FOR HCVP **/
    if(! U->predstride) {
        // Only consult HCVP if E-Stride did not generate a prediction.
        auto corr_prediction = VP.getPrediction(seq_no, inst_info);
        if(std::get<0>(corr_prediction)) {
            predicted_value = std::get<1>(corr_prediction);
            U->predhcvp = true;
            return true;
        }
    }
    /** END FOR HCVP **/

    return (U->predstride || U->predhcvp);
}

void updatePredictor (uint64_t seq_no, uint64_t actual_addr, uint64_t actual_value, uint64_t actual_latency)
{
    (void) actual_addr;

    ForUpdate *U;
    U = &Update[seq_no & (MAXINFLIGHT - 1)];

    /** BEGIN FOR HCVP **/
    auto seq_no_to_info_search = seq_no_to_info.find(seq_no);
    if(seq_no_to_info_search->second.eligible) {
        VP.updatePredictor(seq_no, actual_value);
    }
    seq_no_to_info.erase(seq_no_to_info_search);
    /** END FOR HCVP **/

    if (U->todo == 1) {
        //just to force allocations and update on both predictors
        U->prediction_result = 0;
        UpdateStridePred (U, actual_value, (int) actual_latency);
        U->todo = 0;
    }
    seq_commit = seq_no;
}

void
speculativeUpdate (uint64_t seq_no,    // dynamic micro-instruction # (starts at 0 and increments indefinitely)
           bool eligible,    // true: instruction is eligible for value prediction. false: not eligible.
           uint8_t prediction_result,    // 0: incorrect, 1: correct, 2: unknown (not revealed)
           // Note: can assemble local and global branch history using pc, next_pc, and insn.
           uint64_t
           pc, uint64_t next_pc, InstClass insn, uint8_t piece,
           // Note: up to 3 logical source register specifiers, up to 1 logical destination register specifier.
           // 0xdeadbeef means that logical register does not exist.
           // May use this information to reconstruct architectural register file state (using log. reg. and value at updatePredictor()).
           uint64_t src1, uint64_t src2, uint64_t src3, uint64_t dst)
{
    (void) piece;
    (void) dst;

    /** BEGIN FOR HCVP **/
    auto seq_no_to_info_search = seq_no_to_info.find(seq_no);
    seq_no_to_info_search->second.type = insn;
    seq_no_to_info_search->second.eligible = eligible;

    if((insn == condBranchInstClass) || (insn == uncondIndirectBranchInstClass) || (insn == uncondDirectBranchInstClass)) {
        int outcome = 0;
        if((pc + 4) != next_pc) {
            outcome = 1;
        }

        VP.shiftBHR(outcome);
    }
    /** END FOR HCVP **/

    // the framework does not really allow  to filter the predictions, so we predict every instruction
    ForUpdate *U;
    U = &Update[seq_no & (MAXINFLIGHT - 1)];

    LastMispVT++;

    if (eligible) {
        U->NbOperand = (src1 != 0xdeadbeef) + (src2 != 0xdeadbeef) + (src3 != 0xdeadbeef);
        U->todo = 1;
        U->INSTTYPE = insn;
        U->prediction_result = (prediction_result == 1);
        if (SafeStride < (1 << 15) - 1)
            SafeStride++;
        if (prediction_result != 2) {
            if (prediction_result) {
                if (U->predstride)
                    if (SafeStride < (1 << 15) - 1)
                        SafeStride += 4 * (1 + (insn == loadInstClass));
            } else {
                if (U->predstride)
                    SafeStride -= 1024;
            }
        }
    }

    bool isCondBr = insn == condBranchInstClass;
    bool isUnCondBr = insn == uncondIndirectBranchInstClass || insn == uncondDirectBranchInstClass;
    //path history
    // just to have a longer history without (software) management
    if ((isCondBr) || (isUnCondBr)) {
        if (pc != next_pc - 4) {
            for (int i = 7; i > 0; i--)
                gpath[i] = (gpath[i] << 1) ^ ((gpath[i - 1] >> 63) & 1);
            gpath[0] = (gpath[0] << 1) ^ (pc >> 2);
            gtargeth = (gtargeth << 1) ^ (next_pc >> 2);
        }
    }
}

void beginPredictor (int argc_other, char **argv_other)
{
    (void) argc_other;
    (void) argv_other;

    /** BEGIN FOR HCVP **/
    VP.init(128, 10);
    /** END FOR HCVP **/
}

void endPredictor ()
{
}

void getPredStride (ForUpdate * U, uint64_t & predicted_value, uint64_t seq_no)
{
    bool predstride = false;

    int B[NBWAYSTR];
    int TAG[NBWAYSTR];
    uint64_t pc = U->pc;
    //use a 3-way skewed-associative structure  for the stride predictor
    for (int i = 0; i < NBWAYSTR; i++) {
        //B[i] index in way i ; TAG[i] tag in way i;
        B[i] = ((((pc) ^ (pc >> (2 * LOGSTR - i)) ^ (pc >> (LOGSTR - i)) ^ (pc >> (3 * LOGSTR - i))) * NBWAYSTR) + i) % (NBWAYSTR * (1 << LOGSTR));
        int j = (NBWAYSTR - i);
        if (j < 0)
            j = 0;

        TAG[i] = ((pc >> (LOGSTR - j)) ^ (pc >> (2 * LOGSTR - j)) ^ (pc >> (3 * LOGSTR - j)) ^ (pc >> (4 * LOGSTR - j))) & ((1 << TAGWIDTHSTR) - 1);
        U->B[i] = B[i];
        U->TAGSTR[i] = TAG[i];
    }

    int STHIT = -1;
    for (int i = 0; i < NBWAYSTR; i++) {
        if (STR[B[i]].tag == TAG[i]) {
            STHIT = B[i];
            break;
        }
    }
    U->STHIT = STHIT;
    if (STHIT >= 0)
        if (SafeStride >= 0) {                // hit
            uint64_t LastCommitedValue = STR[STHIT].LastValue;

            if (STR[STHIT].conf >= MAXCONFIDSTR / 4) {
                int inflight = 0;
                // compute the number of inflight instances of the instruction
                for (uint64_t i = seq_commit + 1; i < seq_no; i++) {
                    inflight += (Update[i & (MAXINFLIGHT - 1)].pc == pc);
                }
                predicted_value = (uint64_t) ((int64_t) LastCommitedValue + ((inflight + 1) * ((int64_t) STR[STHIT].Stride)));
                predstride = true;
            }
        }
    U->predstride = predstride;
}

// Update of the Stride predictor
// function determining whether to  update or not confidence on a correct prediction
bool strideupdateconf (ForUpdate * U, uint64_t actual_value, int actual_latency, int stride)
{
#define UPDATECONFSTR2 (((!U->prediction_result) || (U->predstride)) && ((random () & ((1 << (NOTLLCMISS + NOTL2MISS + NOTL1MISS + 2*MFASTINST  + 2*(U->INSTTYPE!=loadInstClass))) - 1)) == 0))
#define UPDATECONFSTR1 (abs (stride >= 8) ? (UPDATECONFSTR2 || UPDATECONFSTR2) : (UPDATECONFSTR2))
#define UPDATECONFSTR (abs (stride >= 64) ? (UPDATECONFSTR1 || UPDATECONFSTR1) : (UPDATECONFSTR1))
    return (UPDATECONFSTR & ((abs (stride) > 1) || (U->INSTTYPE != loadInstClass) || ((stride == -1) & ((random () & 1) == 0)) || ((stride == 1) & ((random () & 3) == 0))));
//All strides are not equal: the smaller the stride the smaller the benefit (not huge :-))

}

//Allocate or not if instruction absent from the predictor
bool StrideAllocateOrNot (ForUpdate * U, uint64_t actual_value, int actual_latency)
{
    bool X = false;
#define LOGPBINVSTR 4
    switch (U->INSTTYPE) {
    case aluInstClass:
    case storeInstClass:
        X = ((random () & ((1 << (LOGPBINVSTR + 2)) - 1)) == 0);
        break;
    case fpInstClass:
        X = ((random () & ((1 << (LOGPBINVSTR)) - 1)) == 0);
        break;
    case slowAluInstClass:
        X = ((random () & ((1 << (LOGPBINVSTR)) - 1)) == 0);
        break;
    case loadInstClass:
        X =
            ((random () &
              ((1 << (NOTLLCMISS + NOTL2MISS + NOTL1MISS + MFASTINST)) - 1)) ==
             0);
        break;
    };

    return (X);
}

void UpdateStridePred (ForUpdate * U, uint64_t actual_value, int actual_latency)
{
    int B[NBWAYSTR];
    int TAG[NBWAYSTR];
    for (int i = 0; i < NBWAYSTR; i++) {
        B[i] = U->B[i];
        TAG[i] = U->TAGSTR[i];
    }
    int STHIT = -1;

    for (int i = 0; i < NBWAYSTR; i++) {
        if (STR[B[i]].tag == TAG[i]) {
            STHIT = B[i];
            break;
        }
    }

    if (STHIT >= 0) {
        uint64_t LastValue = STR[STHIT].LastValue;
        uint64_t Value = (uint64_t) ((int64_t) LastValue + (int64_t) STR[STHIT].Stride);
        int64_t INTER = abs (2 * ((int64_t) actual_value - (int64_t) LastValue) - 1);

        uint64_t stridetoalloc = (INTER < (1 << LOGSTRIDE)) ? (uint64_t) ((int64_t) actual_value - (int64_t) LastValue) : 0;

        STR[STHIT].LastValue = actual_value;

        //special case when the stride is not determined
        if (STR[STHIT].NotFirstOcc > 0) {
            if (Value == actual_value) {
                if (STR[STHIT].conf < MAXCONFIDSTR) {
                    if (strideupdateconf (U, actual_value, actual_latency, (int) stridetoalloc))
                        STR[STHIT].conf++;
                }

                if (STR[STHIT].u < 3)
                    if (strideupdateconf (U, actual_value, actual_latency, (int) stridetoalloc))
                        STR[STHIT].u++;
                if (STR[STHIT].conf >= MAXCONFIDSTR / 4)
                    STR[STHIT].u = 3;
            } else {
                //misprediction
                {
                    if (STR[STHIT].conf > (1 << (WIDTHCONFIDSTR - 3))) {
                        STR[STHIT].conf -= (1 << (WIDTHCONFIDSTR - 3));
                    } else {
                        STR[STHIT].conf = 0;
                        STR[STHIT].u = 0;
                    }
                }

                // this allows to  restart a new sequence with a different   stride
                STR[STHIT].NotFirstOcc = 0;
            }
        } else {
            //First occurence
            //        if (STR[STHIT].NotFirstOcc == 0)
            if (stridetoalloc != 0)
                //             if ((stridetoalloc != 0) && (stridetoalloc!=1)  && (((int64_t) stridetoalloc) != -1))
                // we do not waste the stride predictor storage for stride zero
            {
                STR[STHIT].Stride = stridetoalloc;
            } else {
                // do not pollute the stride predictor with constant data or with invalid strides
                STR[STHIT].Stride = 0xffff;
                STR[STHIT].conf = 0;
                STR[STHIT].u = 0;
            }

            STR[STHIT].NotFirstOcc++;
        }
    } else                // the address was not present and is not predicted by VTAGE
        if (!U->prediction_result) {
            if (StrideAllocateOrNot (U, actual_value, actual_latency)) {            // try to allocate
                int X = random () % NBWAYSTR;
                bool done = false;
                // the target entry is not a stride candidate
                for (int i = 0; i < NBWAYSTR; i++) {
                    STHIT = B[X];
                    if (STR[STHIT].conf == 0) {
                        STR[STHIT].conf = 1;    //just to allow not to ejected before testing if possible stride candidate
                        STR[STHIT].u = 0;
                        STR[STHIT].tag = TAG[X];
                        STR[STHIT].Stride = 0;
                        STR[STHIT].NotFirstOcc = 0;
                        STR[STHIT].LastValue = actual_value;
                        done = true;
                        break;
                    }
                    X = (X + 1) % NBWAYSTR;
                }
                // the target entry has not been useful recently
                if (!done)
                    for (int i = 0; i < NBWAYSTR; i++) {
                        STHIT = B[X];
                        if (STR[STHIT].u == 0) {
                            STR[STHIT].conf = 1;
                            STR[STHIT].u = 0;
                            STR[STHIT].tag = TAG[X];
                            STR[STHIT].Stride = 0;
                            STR[STHIT].NotFirstOcc = 0;
                            STR[STHIT].LastValue = actual_value;
                            done = true;
                            break;

                        }
                        X = (X + 1) % NBWAYSTR;
                    }
                //if unable to allocate: age some target entry
                if (!done) {
                    if ((random () & ((1 << (2 + 2 * (STR[STHIT].conf > (MAXCONFIDSTR) / 8) + 2 * (STR[STHIT].conf >= MAXCONFIDSTR / 4))) - 1)) == 0)
                        STR[STHIT].u--;
                }

            }
        }
    //////////////////////////////
}

